module Gale(
    GaleContext,
    galeNextPuff, 
    newGaleContext, 
    reconnectGaleContext, 
    dumpKey, 
    galeSendPuff, 
    destroyGaleContext,
    hostStrings,
    galeLastDitchEffortPuff, 
    galeWillPuff, 
    getGaleDir) where


import Char(chr,ord)
import IO hiding(bracket, bracket_)
import List
import Monad(when)
import System(getEnv)
import Time
import Maybe

import Network
import Data.Word(Word8, Word32)
import Control.Exception
import Control.Concurrent
import Data.Bits

import ErrorLog
import GenUtil
import Puff
import RSA
import SimpleParser
import EIO
import SHA1


-- TODO - prove concurrent-correctness, make sure all network errors are accounted for.

-------------------
-- Gale Constants
-------------------

galePort :: PortNumber
galePort = 11512
hostStrings s = [s, "gale." ++ s, s ++ ".gale.org."]

galeRSAModulusBits = 1024
galeRSAModulusLen = (galeRSAModulusBits + 7) `div` 8
galeRSAPrimeBits = (galeRSAModulusBits + 1) `div` 2
galeRSAPrimeLen = (galeRSAPrimeBits + 7) `div` 8

data Connection = Connection {cThread :: ThreadId, cHandle :: !(MVar Handle), cCategory :: [Category]}
data GaleContext = GaleContext { channel :: Chan Puff, proxy :: !(MVar [String]), mainConnection :: Connection, keyConnection :: !(MVar Connection), keyCache :: !(MVar [(Key,EvpPkey)]) }
type PuffStatus = ()


void a = a >> return ()


getGaleDir :: IO String
getGaleDir = do
    gd <- lookupEnv "GALE_DIR"
    case gd of
	Just v -> return $ v ++ "/"
	Nothing -> do
	    h <- getEnv "HOME"
	    return (h ++ "/.gale/")
	    
	
    

-----------------
-- Implementation
-----------------

newGaleContext ps cs = do
    cats <- tryMap catParseNew cs
    c <- newChan
    keychan <- newChan
    pmv <- newMVar ps
    mc <- createConnection (if null ps then nub (concatMap (hostStrings . snd) cats) else ps) cats c
    kc <- createConnection (if null ps then nub (concatMap (hostStrings . snd) cats) else ps) [] keychan
    keycachev <- newMVar []
    kcv <- newMVar kc
    let gc = GaleContext { channel = c, proxy = pmv, mainConnection = mc, keyConnection = kcv, keyCache = keycachev}
    return gc


destroyGaleContext gc = do
    destroyConnection (mainConnection gc)
    kc <- takeMVar (keyConnection gc)
    destroyConnection kc




destroyConnection :: Connection -> IO ()
destroyConnection con  = do
    killThread (cThread con)

createConnection :: [String] -> [Category] -> Chan Puff -> IO Connection
createConnection proxys cats channel = do
    hv <- newEmptyMVar
    thd <- forkIO (connectThread channel cats proxys hv)
    return $ Connection {cHandle = hv, cCategory = cats, cThread = thd}

reconnectConnection :: Connection -> [String] -> Chan Puff -> IO ()
reconnectConnection c _ _ = void $ forkIO $ attempt $ do
    h <- readMVar (cHandle c)
    hClose h



attemptConnect s = do
    h <- connectTo s (PortNumber galePort)
    hSetBuffering h NoBuffering
    return h

spc [] = []
spc s = v : spc (drop 1 r) where
    (v,r) = span (/= ':') s

emptyPuffer :: MVar Handle -> IO ()
emptyPuffer hv = repeatM_ (threadDelay 30000000 >> sendEmptyPuff) where
	sendEmptyPuff = withMVar hv $ \h -> do
		putWord32 h 0
		putWord32 h 8
		putWord32 h 0
		putWord32 h 0
		hFlush h

connectThread :: Chan Puff -> [Category] -> [String] -> MVar Handle -> IO ()
connectThread c cs ds hv = retry 5.0 ("ConnectionError " ++ unwords ds) doit where
    openHandle = trySeveral (map attemptConnect ds)
    doit = bracket openHandle hClose $ \h -> do
	putWord32 h 1
	v <- readWord32 h
	let gs = concatInter ":" (map catShowOld cs)
	putWord32 h 2
	putWord32 h (fromIntegral $ length gs * 2)
	putRaw h $ galeEncodeString gs
	hFlush h
	bracket_ (putMVar hv h) (takeMVar hv) $
	  bracket (forkIO (emptyPuffer hv)) killThread $ \_ -> repeatM_ $ do
	    w <- readWord32 h
	    l <- readWord32 h
	    d <- readRaw h (fromIntegral l)
	    when (w == 0) $ do
		let (l,d') = xdrReadUInt d
		cat <- tryMapM parseCategoryOld (spc $ galeDecodeString (take (fromIntegral $ l ) d'))
		ct <- getClockTime
		let ef = \xs -> (("_ginsu.timestamp",FragmentTime ct):("_ginsu.spumbuster", FragmentText (sha1ShowHash $ sha1 d)):xs)
		writeChan c $ Puff { signature = [], cats = cat, fragments = (ef (decodeFragments $ drop (fromIntegral l + 4) d'))}

galeNextPuff :: GaleContext -> IO Puff
galeNextPuff gc = do
    p <- readChan $ channel gc
    p' <- galeDecryptPuff gc p
    putLog LogDebug $ "Puff gotten: \n" ++ (indent 4 $ showPuff p')
    return p'



maybeModifyMVar :: MVar a -> (a -> IO a) -> IO ()
maybeModifyMVar mv fn = do
    v <- tryTakeMVar mv
    case v of
	    Just v -> handle (\e -> putMVar mv v >> throw e) (fn v) >>= putMVar mv
	    Nothing -> return ()


reconnectGaleContext gc = do
    p <- readMVar $ proxy gc
    reconnectConnection (mainConnection gc) p (channel gc)
    maybeModifyMVar (keyConnection gc) (\c -> reconnectConnection c p (channel gc) >> return c)



galeSendPuff :: GaleContext -> Puff -> IO PuffStatus
galeSendPuff gc puff = do
    putLog LogInfo $ "sending puff:\n" ++ (indent 4 $ showPuff puff)
    kc <- readMVar (keyConnection gc)
    d <- createPuff  gc False puff
    retry 3.0 "error sending puff" $ withMVar (cHandle kc) $ \h -> putRaw h d >> hFlush h
    --sendPuff kc puff

galeWillPuff :: GaleContext -> Puff -> IO ()
galeWillPuff gc puff = do
    putLog LogDebug $ "willing puff:\n" ++ (indent 4 $ showPuff puff)
    kc <- readMVar (keyConnection gc)
    d <- createPuff gc True puff
    retry 3.0 "error sending puff" $ withMVar (cHandle kc) $ \h -> putRaw h d >> hFlush h
    --sendPuff kc puff

-- for emergencies.
galeLastDitchEffortPuff :: Puff -> IO Bool
galeLastDitchEffortPuff p = undefined

createPuff :: GaleContext -> Bool -> Puff -> IO [Word8]
createPuff gc will p@(Puff {signature = ((Signed k@(Key kn _)):sr)}) =
    getPrivateKey gc kn >>= \v -> case v of
	Nothing -> createPuff gc will $ p {signature = sr}
	Just (Key _ kfl,pkey) -> do
	    sfl <- case fragmentString "key.owner" kfl of
		Just o -> return [("message/sender", FragmentText o)]
		Nothing -> return []
	    let fl = xdrWriteUInt 0 (createFragments (fragments p `mergeFrags` sfl))
	    sig <- signAll pkey fl
	    let sd = signature_magic1 ++ xdrWriteUInt (fromIntegral $ length sig) [] ++ sig ++ pubkey_magic3 ++ xdrWriteUInt (fromIntegral $ length kn) [] ++ galeEncodeString kn
	    let fd = xdrWriteUInt (fromIntegral $ length sd) $ sd ++ fl
	    createPuff gc will $ p {signature = sr, fragments = [("security/signature",FragmentData fd)]}
createPuff _ will puff = do
    let cn = galeEncodeString (concatInter ":" (map catShowOld $ cats puff))
    let ad = xdrWriteUInt (fromIntegral $ length cn) (cn ++ xdrWriteUInt 0 []) ++ createFragments (fragments puff)
    let pd = xdrWriteUInt (if will then 1 else 0) (xdrWriteUInt (fromIntegral $ length ad) ad)
    evaluate pd



createFragments :: [(String,Fragment)] -> [Word8]
createFragments fl = concatMap f fl where
    f (s,f) = xdrWriteUInt t $ xdrWriteUInt (fromIntegral $ length n + length xs) $ n ++ xs where
	(t, xs) = g f
	n = xdrWriteUInt (fromIntegral $ length s) (galeEncodeString s)
    g (FragmentData ws) = (1, ws)
    g (FragmentText s) = (0, galeEncodeString s)
    g (FragmentTime (TOD s _)) = (2, replicate 4 0 ++ xdrWriteUInt (fromIntegral s) (replicate 8 0))
    g (FragmentInt i) = (3, xdrWriteUInt (fromIntegral i) [])
    g (FragmentNest fl) = (4, createFragments fl)





--------------------
-- Security routines
--------------------


unsignData :: [Word8] -> IO ([(String,Fragment)],Key)
unsignData xs = do
    key <- parseKey $ take (fromIntegral (l - (8 + sb))) (drop (fromIntegral sb) xs'')
    return (fl,key) where
	(l,xs') = xdrReadUInt xs
	(sb,xs'') = xdrReadUInt (drop 4 xs')
	fl = (decodeFragments $ drop (fromIntegral l + 4) xs')

unsignFragments :: [(String, Fragment)] -> IO [(String, Fragment)]
unsignFragments tfl | (xs:_) <- [xs | ("security/signature",FragmentData xs) <- tfl] = do
    (fl,_) <- unsignData xs
    unsignFragments fl
unsignFragments x = return x

galeDecryptPuff :: GaleContext -> Puff -> IO Puff
galeDecryptPuff gc p | (Just xs) <- getFragmentData p "security/signature" = tryElse p $ do
    let (l,xs') = xdrReadUInt xs
	(sb,xs'') = xdrReadUInt (drop 4 xs')
	fl = (decodeFragments $ drop (fromIntegral l + 4) xs') ++ [f|f <- fragments p, fst f /= "security/signature"]
    key <- parseKey $ take (fromIntegral (l - (8 + sb))) (drop (fromIntegral sb) xs'')
    galeDecryptPuff gc $ p {signature = (Unverifyable key): signature p, fragments = fl} 
galeDecryptPuff gc p | (Just xs) <- getFragmentData p "security/encryption" = tryElse p $ do
    (cd,ks) <- parser pe xs
    dfl <- first (map (td' cd) ks) 
    let dfl' = dfl ++ [f|f <- fragments p, fst f /= "security/encryption"]
    galeDecryptPuff gc $ p {signature = (Encrypted (map (\(_,n,_) -> n) ks)): signature p, fragments = dfl'}  where
	pe = (parseExact cipher_magic1 >> pr parseNullString) <|> (parseExact cipher_magic2 >> pr parseLenString)
	pk pkname iv = do
	    kname <- pkname
	    keydata <- parseLenData
	    return $ (iv,kname,keydata)
	pr pkname = do
	    iv <- parseSome 8
	    keycount <-  parseIntegral32
	    ks <- replicateM keycount (pk pkname iv)
	    xs <- parseRest
	    return (xs,ks)
	td' cd (iv,kname,keydata) = do
	    Just (_,pkey) <- getPrivateKey gc kname
	    dd <- decryptAll keydata iv pkey cd
	    let dfl = decodeFragments (drop 4 dd)
	    return dfl
galeDecryptPuff gc x = return x


getPrivateKey :: GaleContext -> String -> IO (Maybe (Key,EvpPkey))
getPrivateKey gc kn = modifyMVar (keyCache gc) f where
    f kc = case [x|x@(Key kn' _,_) <- kc, kn == kn'] of
	(v:_) -> return (kc,Just v)
	[] -> do
	    v <- ioMp getFromDisk
	    return ((maybeToList v ++ kc),v)
    getFromDisk = do
	gd <- getGaleDir
	let p = (gd ++ "/auth/private/")
	    knames = [p ++ kn, p ++ kn ++ ".gpri"]
	    gn = (map (\fn -> (getRawContents fn >>= \c -> return (c,fn))) knames)
	(key_c, fn) <- (first  gn) 
	key <- parseKey key_c
	rsa <- privkeyToRSA key
	pkey <- pkeyNewRSA rsa
	putLog LogNotice $ "Retrieved private key from disk: " ++ fn
	return (key,pkey)



----------------------
-- Gale stream Parsing
----------------------

parseWord32 :: GenParser Word8 Word32
parseWord32 = do
    [b1,b2,b3,b4] <- parseSome 4
    return $ (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|.
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)

parseLenData = parseWord32 >>= parseSome

parseIntegral32 :: Integral a => GenParser Word8 a
parseIntegral32 = fmap fromIntegral parseWord32

parseLenString :: GenParser Word8 String
parseLenString = do
    len <- parseIntegral32
    w <- parseSome (len * 2)
    return $ galeDecodeString w

parseNullString :: GenParser Word8 String
parseNullString = (sat (== 0) >> return "") <|> (parseSome 1 >>= \[x] -> fmap ((chr $ fromIntegral x):) parseNullString)

galeDecodeString :: [Word8] -> String
galeDecodeString [] = []
galeDecodeString (b1:b2:xs) = (chr $ (fromIntegral b2) .|. (fromIntegral b1 `shiftL` 8)) : galeDecodeString xs
galeDecodeString _ = error "invalid gale string"


xdrWriteUInt :: Word32 -> [Word8] -> [Word8]
xdrWriteUInt x bs = (b1:b2:b3:b4:bs) where
    b1 = fromIntegral $ (x `shiftR` 24) .&. 0xFF
    b2 = fromIntegral $ (x `shiftR` 16) .&. 0xFF
    b3 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
    b4 = fromIntegral $ x .&. 0xFF

xdrReadUInt :: [Word8] -> ( Word32,[Word8])
xdrReadUInt (b1:b2:b3:b4:bs) = (x,bs) where
    x = (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|.
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)
xdrReadUInt bs = error $ "xdrReadUInt " ++ show bs

putWord32 h x = do
    putRaw h $ xdrWriteUInt x []

readWord32 h = do
    xs <- readRaw h 4
    return $ fst $ xdrReadUInt xs

galeEncodeString :: String -> [Word8]
galeEncodeString cs = concatMap (f . ord) (concat $ map (\c -> if c == '\n' then "\r\n" else [c]) cs) where
    f x = (b1:b2:[]) where
	b1 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
	b2 = fromIntegral $ x .&. 0xFF

privkeyToRSA :: Key -> IO RSA
privkeyToRSA (Key _ fl)  = do
    rsa <- rsaNew
    fragmentData "rsa.modulus" fl >>= bn_bin2bn >>= rsaSetN rsa
    fragmentData "rsa.exponent" fl >>= bn_bin2bn >>= rsaSetE rsa
    fragmentData "rsa.private.exponent" fl >>= bn_bin2bn >>= rsaSetD rsa
    fragmentData "rsa.private.coefficient" fl >>= bn_bin2bn >>= rsaSetIQMP rsa
    xs <- fragmentData "rsa.private.prime" fl
    let (p,q) = splitAt galeRSAPrimeLen xs
    bn_bin2bn p >>= rsaSetP rsa
    bn_bin2bn q >>= rsaSetQ rsa
    xs <- fragmentData "rsa.private.prime.exponent" fl
    let (dmp1,dmq1) = splitAt galeRSAPrimeLen xs
    bn_bin2bn dmp1 >>= rsaSetDMP1 rsa
    bn_bin2bn dmq1 >>= rsaSetDMQ1 rsa
    rsaCheckKey rsa
    return rsa


--key_response_cat = "_gale.key."
--key_request_cat = "_gale.query."

--------------
-- Key Parsing
--------------

stons :: [Char] -> [Word8]
stons = map (fromIntegral . ord)

private_magic1, private_magic2, private_magic3 :: [Word8]
cipher_magic1, cipher_magic2 :: [Word8]
pubkey_magic1, pubkey_magic2, pubkey_magic3 :: [Word8]
signature_magic1 :: [Word8]

cipher_magic1 = stons "h\DC3\002\000"
cipher_magic2 = stons "h\DC3\002\001"

signature_magic1 = stons "h\DC3\001\000"

private_magic1 = stons "h\DC3\000\001"
private_magic2 = stons "h\DC3\000\003"
private_magic3 = stons "GALE\000\002"


pubkey_magic1 = stons "h\DC3\000\000"
pubkey_magic2 = stons "h\DC3\000\002"
pubkey_magic3 = stons "GALE\000\001"


flipLocalPart :: String -> String
flipLocalPart s | '@' `notElem` s = s
flipLocalPart s = nbp ++ ep where
    nbp = concat $ reverse (groupBy f bp)
    (bp,ep) = span (/= '@') s
    f '.' '.' = True
    f x y | x /= '.' && y /= '.' = True
    f _ _ = False

getRawContents :: String -> IO [Word8]
getRawContents fn = do
    c <- readFile fn
    return (map (fromIntegral . ord) c)


keyDecode12 xs = fl where
    fl = [("rsa.modulus",FragmentData modulusD),
	  ("rsa.exponent",FragmentData exponentD),
	  ("rsa.private.exponent",FragmentData privateExponentD),
	  ("rsa.private.prime",FragmentData privatePrimeD),
	  ("rsa.private.prime.exponent",FragmentData privatePrimeExponentD),
	  ("rsa.private.coefficient",FragmentData privateCoefficientD),
	  ("rsa.bits", FragmentInt (fromIntegral bits))
	 ]
    (bits,ys) = xdrReadUInt xs
    (modulusD,xs') = splitRLE galeRSAModulusLen ys
    (exponentD,xs'') = splitRLE galeRSAModulusLen xs'
    (privateExponentD,xs''') = splitRLE (galeRSAPrimeLen * 2) xs''
    (privatePrimeD,xs'''') = splitRLE galeRSAModulusLen xs'''
    (privatePrimeExponentD,xs''''') = splitRLE (galeRSAPrimeLen * 2) xs''''
    (privateCoefficientD,_) = splitRLE galeRSAPrimeLen xs'''''



keyParse :: GenParser Word8 Key
keyParse = choice [ppk1,ppk2,ppk3,pk1,pk2,pk3] where
    ppk1 = do
	parseExact private_magic1
	kn <- fmap flipLocalPart parseNullString
	fl <- toParser keyDecode12
	return $ Key kn fl
    ppk2 = do
	parseExact private_magic2
	kn <- fmap flipLocalPart parseLenString
	fl <- toParser keyDecode12
	return $ Key kn fl
    ppk3 = do
	parseExact private_magic3
	kn <- fmap flipLocalPart parseLenString
	fl <- toParser decodeFragments
	return $ Key kn fl
    pk1 = do
	parseExact pubkey_magic1
	kn <- parseNullString
	(eof >> return (Key kn [])) <|> do
	comment <- parseNullString
	fl <- parsePublic12
	signature <- parseRest
	return $ Key kn ([("key.owner", FragmentText comment)] ++ fl)
    pk2 = do
	parseExact pubkey_magic2
	kn <- parseLenString
	(eof >> return (Key kn [])) <|> do
	comment <- parseLenString
	fl' <- parsePublic12
	let fl = (("key.owner", FragmentText comment):fl')
	(eof >> return (Key kn fl)) <|> do
	ts <- parseSome 16
	te <- parseSome 16
	signature <- parseRest
	return $ Key kn $ fl ++ [("key.signed",FragmentTime (decodeTime ts)), ("key.expires",FragmentTime (decodeTime te))]
    pk3 = do
	parseExact pubkey_magic3
	kn <- parseLenString
	fl <- toParser decodeFragments
	return $ Key kn fl
    parsePublic12 = do
	bits <- parseIntegral32
	modulus <- (MkP (\x -> Just (splitRLE galeRSAModulusLen x)))
	exponent <- (MkP (\x -> Just (splitRLE galeRSAModulusLen x)))
	return [("rsa.modulus", FragmentData modulus),
		("rsa.exponent", FragmentData exponent),
		("rsa.bits", FragmentInt bits)]

parseKey :: [Word8] -> IO Key
parseKey xs = do
    (Key kn fl) <- parser keyParse xs
    fl <- unsignFragments fl
    return $ Key kn fl

showKey :: Key -> String
showKey (Key n fl) = "Key: " ++ n ++ "\n" ++ (indentLines 4 $ showFragments fl)

---------------
-- data parsing
---------------

splitRLE :: Int -> [Word8] -> ([Word8], [Word8])
splitRLE n _ | n < 0 = error "invalid RLE encoding"
splitRLE 0 xs = ([],xs)
splitRLE n (c:xs) | c .&. 0x80 /= 0 = (take count xs ++ ys,rest) where
    count = fromIntegral $ (c .&. 0x7f) + 1
    (ys,rest) = splitRLE (n - count) (drop count xs)
splitRLE n (c:x:xs) | c .&. 0x80 == 0 = (replicate count x ++ ys,rest) where
    count = fromIntegral $ (c .&. 0x7f) + 1
    (ys,rest) = splitRLE (n - count) xs
splitRLE _ _  = error "invalid RLE encoding"

toParser :: ([c] -> a) -> GenParser c a
toParser f = MkP (\xs -> Just (f xs, []))

decodeFragments :: [Word8] -> [(String,Fragment)]
decodeFragments [] = []
decodeFragments xs = decodeFragment ft f : decodeFragments r where
    (ft,xs') = xdrReadUInt xs
    (fl,xs'') = xdrReadUInt xs'
    (f,r) = splitAt (fromIntegral fl) xs''

decodeFragment :: Word32 -> [Word8] -> (String, Fragment)
decodeFragment t xs = (fn,f t) where
    (fnl,xs') = xdrReadUInt xs
    (fn,xs'') = liftT2 (galeDecodeString,id) (splitAt (fromIntegral $ fnl * 2) xs')
    f 0 = FragmentText (galeDecodeString xs'')
    f 1 = FragmentData xs''
    f 2 = FragmentTime $ decodeTime xs''
    f 3 = FragmentInt (fromIntegral $ fst $ xdrReadUInt xs'')
    f 4 = FragmentNest (decodeFragments xs'')
    f x = error $ "unknown fragment: " ++ show x

decodeTime :: [Word8] -> ClockTime
decodeTime xs = TOD (fromIntegral t) 0 where
    (t,_) = xdrReadUInt (drop 4 xs)

{-
decodeTime (x:y:_) | x > 0 || y > 0 = endOfTime
decodeTime xs = TimeDiff {tdDay = 0, tdPicosec = 0, tdYear = 0, tdMonth = 0, tdHour = 0, tdMin = fromIntegral (t `div` 60) , tdSec = fromIntegral (t `mod` 60)} `addToClockTime` epoch where
    (t,_) = xdrReadUInt (drop 4 xs)

-}


----------------
-- test code
----------------

dumpKey :: String -> IO ()
dumpKey arg = do
    c <- getRawContents arg
    key <- parseKey c
    putStrLn $ showKey key


