module Zlib (
    Stream, -- data Stream
    
    defaultCompression, -- :: Int
    
    Flush(..), -- data Flush
               --     = NoFlush
               --     | SyncFlush
               --     | FullFlush
               --     | Finish
    
    Status(..), -- data Status
                --     = Ok        Int [Word8]
                --     | StreamEnd     [Word8]
                --     | NeedDict  Word32
                --     | DataError
    
    deflateInit,  -- :: Int -> IO Stream
    deflateBlock, -- :: Stream -> [Word8] -> Int -> Flush -> IO Status
    deflate,      -- :: Int -> [Word8] -> [Word8]
    
    inflateInit,  -- :: IO Stream
    inflateBlock, -- :: Stream -> [Word8] -> Int -> Flush -> IO Status
    inflate)      -- :: [Word8] -> [Word8]
    
    where

------------------------------------------------------------------------

import Monad     (unless, liftM, liftM2)

import Foreign
import CForeign
import IOExts

#include <zlib.h>

data StreamTag
type Stream = ForeignPtr StreamTag

blockSize :: Int
blockSize = 4096

data Flush = NoFlush | SyncFlush | FullFlush | Finish
data Status = Ok        Int [Word8]
            | StreamEnd     [Word8]
            | NeedDict  Word32
            | DataError

-- Deflation.

defaultCompression :: Int
defaultCompression = #const Z_DEFAULT_COMPRESSION

deflateInit :: Int -> IO Stream
deflateInit level = do
    strmPtr <- mallocBytes (#const sizeof (z_stream))
    (#poke z_stream, zalloc) strmPtr nullFunPtr
    (#poke z_stream, zfree)  strmPtr nullFunPtr
    (#poke z_stream, opaque) strmPtr nullPtr
    ok <- hs_zlib_deflateInit strmPtr (fromIntegral level)
    unless (ok == #const Z_OK) $ do free strmPtr; throwError strmPtr
    newForeignPtr strmPtr $ do deflateEnd strmPtr; free strmPtr

foreign import ccall unsafe hs_zlib_deflateInit :: Ptr StreamTag -> CInt -> IO CInt
#def inline int hs_zlib_deflateInit (z_streamp strm, int level) \
{return deflateInit (strm, level);}

foreign import ccall unsafe deflateEnd :: Ptr StreamTag -> IO CInt

deflateBlock :: Stream -> [Word8] -> Int -> Flush -> IO Status
deflateBlock strm input availOut flush =
    withForeignPtr strm  $ \strmPtr ->
    withArray input      $ \nextIn  ->
    allocaArray availOut $ \nextOut -> do
        let availIn = length input
        (#poke z_stream, next_in)   strmPtr nextIn
        (#poke z_stream, avail_in)  strmPtr (fromIntegral availIn::CUInt)
        (#poke z_stream, next_out)  strmPtr nextOut
        (#poke z_stream, avail_out) strmPtr (fromIntegral availOut::CUInt)
        ok <- deflateC strmPtr $ case flush of
            NoFlush   -> #const Z_NO_FLUSH
            SyncFlush -> #const Z_SYNC_FLUSH
            FullFlush -> #const Z_FULL_FLUSH
            Finish    -> #const Z_FINISH
        let getSizeIn = do
                newAvailIn  <- (#peek z_stream, avail_in)  strmPtr
                return (availIn - fromIntegral (newAvailIn::CUInt))
        let getOutput = do
                newAvailOut <- (#peek z_stream, avail_out) strmPtr
                let sizeOut = availOut - fromIntegral (newAvailOut::CUInt)
                peekArray sizeOut nextOut
        case ok of
            (#const Z_OK)         -> liftM2 Ok        getSizeIn getOutput
            (#const Z_STREAM_END) -> liftM  StreamEnd           getOutput
            _                     -> throwError strmPtr

foreign import ccall unsafe "deflate" deflateC :: Ptr StreamTag -> CInt -> IO CInt

deflate :: Int -> [Word8] -> [Word8]
deflate level input = unsafePerformIO $ do
    strm <- deflateInit level
    deflateLoop strm input

deflateLoop :: Stream -> [Word8] -> IO [Word8]
deflateLoop strm input = do
    let (blockIn, restIn) = splitAt blockSize input
    ok <- deflateBlock strm blockIn blockSize
        (if null restIn then Finish else NoFlush)
    case ok of
        Ok inSize blockOut -> liftM (blockOut++) $ unsafeInterleaveIO $
                                  deflateLoop strm (drop inSize input)
        StreamEnd output   -> return output
        _                  -> withForeignPtr strm throwError

-- Inflation.

inflateInit :: IO Stream
inflateInit = do
    strmPtr <- mallocBytes (#const sizeof (z_stream))
    (#poke z_stream, zalloc)   strmPtr nullFunPtr
    (#poke z_stream, zfree)    strmPtr nullFunPtr
    (#poke z_stream, opaque)   strmPtr nullPtr
    (#poke z_stream, next_in)  strmPtr nullPtr
    (#poke z_stream, avail_in) strmPtr (0::CUInt)
    ok <- hs_zlib_inflateInit strmPtr
    unless (ok == #const Z_OK) $ do free strmPtr; throwError strmPtr
    newForeignPtr strmPtr $ do inflateEnd strmPtr; free strmPtr

foreign import ccall unsafe hs_zlib_inflateInit :: Ptr StreamTag -> IO CInt
#def inline int hs_zlib_inflateInit (z_streamp strm) \
{return inflateInit (strm);}

foreign import ccall unsafe inflateEnd :: Ptr StreamTag -> IO CInt

inflateBlock :: Stream -> [Word8] -> Int -> Flush -> IO Status
inflateBlock strm input availOut flush =
    withForeignPtr strm  $ \strmPtr ->
    withArray input      $ \nextIn  ->
    allocaArray availOut $ \nextOut -> do
        let availIn = length input
        (#poke z_stream, next_in)   strmPtr nextIn
        (#poke z_stream, avail_in)  strmPtr (fromIntegral availIn::CUInt)
        (#poke z_stream, next_out)  strmPtr nextOut
        (#poke z_stream, avail_out) strmPtr (fromIntegral availOut::CUInt)
        ok <- inflateC strmPtr $ case flush of
            NoFlush   -> #const Z_NO_FLUSH
            SyncFlush -> #const Z_SYNC_FLUSH
            FullFlush -> #const Z_FULL_FLUSH
            Finish    -> #const Z_FINISH
        let getSizeIn = do
                newAvailIn  <- (#peek z_stream, avail_in)  strmPtr
                return (availIn - fromIntegral (newAvailIn::CUInt))
        let getOutput = do
                newAvailOut <- (#peek z_stream, avail_out) strmPtr
                let sizeOut = availOut - fromIntegral (newAvailOut::CUInt)
                peekArray sizeOut nextOut
        case ok of
            (#const Z_OK)         -> liftM2 Ok        getSizeIn getOutput
            (#const Z_STREAM_END) -> liftM  StreamEnd           getOutput
            (#const Z_NEED_DICT)  -> liftM  NeedDict  ((#peek z_stream, adler) strmPtr)
            (#const Z_DATA_ERROR) -> return DataError
            _                     -> throwError strmPtr

foreign import ccall unsafe "inflate" inflateC :: Ptr StreamTag -> CInt -> IO CInt

inflate :: [Word8] -> [Word8]
inflate input = unsafePerformIO $ do
    strm <- inflateInit
    inflateLoop strm input

inflateLoop :: Stream -> [Word8] -> IO [Word8]
inflateLoop strm input = do
    let blockIn = take blockSize input
    ok <- inflateBlock strm blockIn blockSize NoFlush
    case ok of
        Ok inSize blockOut -> liftM (blockOut++) $ unsafeInterleaveIO $
                                  inflateLoop strm (drop inSize input)
        StreamEnd output   -> return output
        _                  -> withForeignPtr strm throwError

throwError :: Ptr StreamTag -> IO a
throwError strmPtr = do
    msgPtr <- (#peek z_stream, msg) strmPtr
    msg <- if msgPtr == nullPtr
        then return "zlib error"
        else peekCString msgPtr
    ioError (userError msg)
